import uuid

from qdrant_client import QdrantClient

from vecutils.qdrant import index_vectors, iter_points
from vecutils.utils import chunk

qd = QdrantClient("http://localhost:6333")

points = iter_points(qd, "text2", limit=1000, with_payload=True, with_vectors=True)

for batch in chunk((points), 1000):
    batch = list(batch)
    print(batch[0])
    embeddings = [x.vector["text-embedding-ada-002"] for x in batch]
    docs = [x.payload for x in batch]
    for doc in docs:
        doc["id"] = uuid.uuid4().hex
    index_vectors(qd, "text3", docs, embeddings, vector="text-embedding-ada-002")



import json

from qdrant_client import QdrantClient

from vecutils.qdrant import iter_points

qdrant = QdrantClient("http://129.226.212.197.ipssh.net:36333")


examples = []
for i, points in enumerate(
    iter_points(
        qdrant,
        "similarity_dispatcher_examples",
        with_payload=True,
        with_vectors=True,
        limit=10000,
    ),
):
    examples += [x.model_dump() for x in points]

with open(f"data/points-{i:0>10}.json", mode="w") as f:
    json.dump(examples, f)

questions = [x["payload"]["question"] for x in examples]

with open("data/questions.txt", mode="w") as f:
    f.writelines("\n".join(questions))



import copy
import logging
import uuid
from collections.abc import Awaitable, Callable, Iterable

import tqdm
from qdrant_client import AsyncQdrantClient, QdrantClient, models
from qdrant_client.http.models import PointStruct

from vecutils import batch_create_embeddings
from vecutils.utils import chunk

logger = logging.getLogger(__name__)


def to_dict(x):
    if x is None:
        return {}

    return x.model_dump()


def duplicate_collection(  # noqa: PLR0913
    client: QdrantClient,
    source_colleciton: str,
    target_collection: str,
    delete_vectors: list[str] | None = None,
    add_vectors: dict | None = None,
    indexes: list[dict] | None = None,
    limit: int = 1000,
    timeout: int | None = None,
):
    collection_info = client.get_collection(source_colleciton)
    config = collection_info.config
    logger.info(
        "Duplicating, source: %s, target: %s, vectors_count: %d",
        source_colleciton,
        target_collection,
        collection_info.vectors_count,
    )

    # Update vectors config
    vectors_config = config.params.vectors
    if isinstance(vectors_config, models.VectorParams):
        msg = "Only multi-vector collection is supported"
        raise TypeError(msg)

    if delete_vectors:
        kept_vectors_config = {
            k: v for k, v in vectors_config.items() if k not in delete_vectors
        }
    else:
        kept_vectors_config = vectors_config

    if add_vectors:
        vectors_config = kept_vectors_config | add_vectors

    # Create new collection
    logger.info(
        "Creating collection, name: %s, vectors: %s",
        target_collection,
        list(vectors_config),
    )
    client.create_collection(
        collection_name=target_collection,
        vectors_config=vectors_config,
        shard_number=config.params.shard_number,
        replication_factor=config.params.replication_factor,
        write_consistency_factor=config.params.write_consistency_factor,
        on_disk_payload=config.params.on_disk_payload,
        hnsw_config=to_dict(config.hnsw_config),
        optimizers_config=to_dict(config.optimizer_config),
        wal_config=to_dict(config.wal_config),
        quantization_config=config.quantization_config,
        timeout=timeout,
    )

    # Update payload index
    if indexes:
        indexes = copy.deepcopy(indexes)
        for key, value in collection_info.payload_schema.items():
            indexes += [
                {
                    "field_name": key,
                    "field_type": value.data_type.value,
                },
            ]
        for index in indexes:
            logger.info(
                "Creating payload index, name: %s, schema: %s",
                index["field_name"],
                index["field_type"],
            )
            client.create_payload_index(target_collection, **index)

    # Upserting points
    batches = iter_points(
        client,
        source_colleciton,
        limit=limit,
        with_payload=True,
        with_vectors=list(kept_vectors_config),
    )
    progress_bar = tqdm.tqdm(
        batches,
        total=collection_info.points_count,
        desc="Upserting points",
        leave=False,
    )
    for batch in progress_bar:
        points = [
            models.PointStruct(id=x.id, payload=x.payload, vector=x.vector)
            for x in batch
        ]

        client.upsert(collection_name=target_collection, points=points)
        progress_bar.update(len(points))
    progress_bar.close()


def iter_points(qdrant: QdrantClient, index: str, **kwargs) -> Iterable:
    kwargs.pop("offset", None)
    offset = None

    while True:
        records, offset = qdrant.scroll(collection_name=index, offset=offset, **kwargs)
        yield records

        if not offset:
            break


def get_existed_point_ids(
    qdrant: QdrantClient,
    index: str,
    id_field: str = "id",
    limit: int = 10000,
) -> set[str | int]:
    iterator = iter_points(
        qdrant,
        index,
        limit=limit,
        with_payload=True,
        with_vectors=False,
    )

    return {x.payload[id_field] for batch in iterator for x in batch}


async def index_vectors(
    client: AsyncQdrantClient,
    index: str,
    docs: list[dict],
    embeddings: list[list[float]],
    vector: str | None = None,
) -> None:
    def _get_vector(embedding):
        if not vector:
            return embedding

        return {
            vector: embedding,
        }

    if len(docs) != len(embeddings):
        msg = "mismatched lengths of documents and embeddings: %d != %d"
        raise ValueError(msg, len(docs), len(embeddings))

    points = [
        PointStruct(
            id=uuid.uuid4().hex,
            vector=_get_vector(embedding),
            payload=doc,
        )
        for embedding, doc in zip(embeddings, docs, strict=True)
    ]
    await client.upsert(collection_name=index, wait=True, points=points)


async def batch_index_vectors(  # noqa: PLR0913
    client: AsyncQdrantClient,
    index: tuple[str, str],
    docs: list[dict],
    embedding_fn: Callable[[list[str]], Awaitable[list[list[float]]]],
    format_fn: Callable[[dict], str],
    index_chunk_size: int = 256,
    embedding_batch_size: int = 16,
    embedding_concurrency: int = 10,
) -> None:
    index_name, vector_name = index

    for batch in chunk(docs, size=index_chunk_size):
        batch_texts = list(map(format_fn, batch))
        batch_embeddings = await batch_create_embeddings(
            batch_texts,
            embedding_fn=embedding_fn,
            batch_size=embedding_batch_size,
            concurrency=embedding_concurrency,
        )
        if len(batch_embeddings) != len(batch):
            logger.warning(
                "%d / %d docs failed to create embeddings",
                len(batch) - len(batch_embeddings),
                len(batch),
            )

        if not batch_embeddings:
            continue

        indices, embeddings = zip(*batch_embeddings, strict=True)
        filtered_batch = [batch[i] for i in indices]

        logger.info("update %d embeddings to qdrant", len(indices))
        await index_vectors(client, index_name, filtered_batch, embeddings, vector_name)



import logging
from collections.abc import Awaitable, Callable, Iterable

from vecutils.utils import chunk, gather_with_concurrency

logger = logging.getLogger(__name__)


async def batch_create_embeddings(
    texts: Iterable[str],
    embedding_fn: Callable[[list[str]], Awaitable[list[list[float]]]],
    batch_size: int = 16,
    concurrency: int = 10,
) -> list[tuple[int, list[float]]]:
    async def _create_embeddings(ordered_batch):
        indices, batch = zip(*ordered_batch, strict=True)
        try:
            embeddings = await embedding_fn(batch)

        except Exception:
            logger.exception("failed to create embeddings, texts: %s", batch)

            return False, zip(indices, batch, strict=True)

        return True, list(zip(indices, embeddings, strict=True))

    min_split_batch_size = 2
    ordered_embeddings = []

    futures = []
    for ordered_batch in chunk(enumerate(texts), size=batch_size):
        futures += [_create_embeddings(ordered_batch)]

    while True:
        if not futures:
            break

        responses = await gather_with_concurrency(concurrency, *futures)
        futures = []
        for response in responses:
            success, payload_or_result = response
            if success:
                ordered_embeddings += payload_or_result
                continue

            indices, batch = zip(*payload_or_result, strict=True)
            size = len(indices)
            if size >= min_split_batch_size:
                size //= 2
                futures += [
                    _create_embeddings(zip(indices[:size], batch[:size], strict=True)),
                ]
                futures += [
                    _create_embeddings(zip(indices[size:], batch[size:], strict=True)),
                ]
                continue

    ordered_embeddings.sort(key=lambda x: x[0])

    return ordered_embeddings



import asyncio
import random

from vecutils import Vectorizer, batch_create_embeddings


async def embedding_fn(texts):
    await asyncio.sleep(1)

    if random.random() > 0.1:
        raise ValueError

    return [[i * [1]] for i in range(len(texts))]


async def main():
    texts = [f"{i}" * random.randint(1, 10) for i in range(100)]

    vectroizer = Vectorizer("http://localhost:9394/v1/embeddings", "moka-ai/m3e-base")

    results = await batch_create_embeddings(
        texts,
        embedding_fn=vectroizer.acreate_embeddings,
        batch_size=10,
        concurrency=5,
    )

    print(results)


if __name__ == "__main__":
    asyncio.run(main())



import itertools
import asyncio
from collections.abc import Iterable


def chunk(iterator: Iterable, size: int) -> Iterable[tuple]:
    it = iter(iterator)

    return iter(lambda: tuple(itertools.islice(it, size)), ())


async def gather_with_concurrency(n, *coros, return_exceptions=False):
    semaphore = asyncio.Semaphore(n)

    async def sem_coro(coro):
        async with semaphore:
            return await coro

    return await asyncio.gather(
        *(sem_coro(c) for c in coros),
        return_exceptions=return_exceptions,
    )



import argparse
import logging
import os

from qdrant_client import QdrantClient

from vecutils.qdrant import duplicate_collection


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("--url", type=str, default=None)
    parser.add_argument("--source", type=str, required=True)
    parser.add_argument("--target", type=str, required=True)
    parser.add_argument("--delete-vectors", type=str, nargs="*")
    parser.add_argument("--add-vectors", type=str, nargs="*")
    parser.add_argument("--indexes", type=str, nargs="*")
    parser.add_argument("--chunk-size", type=int, default=500)

    args = parser.parse_args()  # pylint: disable=redefined-outer-name
    if not args.url:
        args.url = os.getenv("QDRANT_URL")

    add_vectors = {}
    if args.add_vectors:
        for spec in args.add_vectors:
            remaining, distance = spec.rsplit("/", 1)
            name, size = remaining.rsplit("/", 1)
            add_vectors[name] = {
                "size": int(size),
                "distance": distance,
            }
    args.add_vectors = add_vectors

    indexes = []
    if args.indexes:
        for spec in args.indexes:
            name, type_ = spec.split("/")
            indexes += [
                {
                    "field_name": name,
                    "field_type": type_,
                },
            ]
    args.indexes = indexes

    return args


def main():
    args = parse_args()

    logging.getLogger("httpx").setLevel(logging.WARNING)
    logging.basicConfig(level=logging.INFO, force=True)

    client = QdrantClient(args.url)

    duplicate_collection(
        client,
        args.source,
        args.target,
        delete_vectors=args.delete_vectors,
        add_vectors=args.add_vectors,
        indexes=args.indexes,
        limit=args.chunk_size,
    )


if __name__ == "__main__":
    main()
